@@ -2,19 +2,47 @@
 The Estimators module contains a variety of statistical estimators that can be applied to multivariate datasets.
 '''
 
-
-# TODO: Consider suggestions at https://stackoverflow.com/questions/35673895/type-hinting-annotation-pep-484-for-numpy-ndarray
-# TODO: Ensure code follows style guide: https://numpydoc.readthedocs.io/en/latest/format.html
 import os
 
-from numba import jit
+import networkx as nx
 import numpy as np
+import tqdm
+from numpy import mean
+from numpy import std as standard_deviation
 from pathos.pools import ProcessPool
 from scipy.stats import rankdata
-import tqdm
 
 
-def kendall_tau(X: np.ndarray, method='A', n_jobs=os.cpu_count()) -> np.float64:
+def pnorm(x: np.ndarray, p=2) -> np.float64:
+    '''
+    Computes the p-norm of a given vector.
+    
+    Parameters
+    ----------
+    x : 1D array-like
+    	An m-dimensional vector.
+    p : float
+    	Order of the norm.
+    
+    Returns
+    -------
+    	: np.float64
+    	P-norm of input vector.
+    
+    References
+    ----------
+    .. [1] "Norm (mathematics).", https://en.wikipedia.org/wiki/Norm_(mathematics)
+    .. [2] "Norm.", https://mathworld.wolfram.com/Norm.html
+    .. [3] "Vector Norm.", https://mathworld.wolfram.com/VectorNorm.html
+    '''
+    result = np.abs(x)
+    result = np.power(result, p)
+    result = np.sum(result)
+    result = np.power(result, 1/p)
+    return result
+
+
+def kendall_tau(X: np.ndarray, method='A', n_jobs=1) -> np.float64:
     '''
     Multivariate Kendall's tau.
 
@@ -36,7 +64,8 @@ def kendall_tau(X: np.ndarray, method='A', n_jobs=os.cpu_count()) -> np.float64:
 
     Raises
     ------
-    NotImplementedError : Method B is not implemented yet.
+    NotImplementedError
+        Method B is not implemented yet.
 
     References
     ----------
@@ -49,21 +78,34 @@ def kendall_tau(X: np.ndarray, method='A', n_jobs=os.cpu_count()) -> np.float64:
     >>> kendall_tau(data)
     1.0
     '''
-    @jit(nopython=True, fastmath=True)
     def scorer(q):
-        score = 0
+        score_q = 0
         for j in range(X.shape[0]):
             if np.all(X[q] < X[j]) or np.all(X[q] > X[j]):
-                score += 1
+                score_q += 1
             elif np.any(X[q] == X[j]):
                 continue
             else:
-                score -= 1
-        return score
+                score_q -= 1
+        return score_q
 
-    pool = ProcessPool(nodes=n_jobs)
-    score = np.sum(pool.map(scorer, range(X.shape[0])))
+    score = 0
+    if n_jobs == 1:
+        for i in tqdm.tqdm(range(X.shape[0])):
+            score += scorer(i)
 
+    else:
+        try:
+            pool = ProcessPool(nodes=n_jobs)
+            score = np.sum(pool.map(scorer, range(X.shape[0])))
+            pool.close()
+            pool.join()
+        except ValueError:
+            pool.restart()
+            pool = ProcessPool(nodes=n_jobs)
+            score = np.sum(pool.map(scorer, range(X.shape[0])))
+            pool.close()
+            pool.join()
     if method == 'A':
         return score / ((X.shape[0] **2 - X.shape[0]))
     elif method == 'B':
@@ -76,24 +118,15 @@ def kendall_tau(X: np.ndarray, method='A', n_jobs=os.cpu_count()) -> np.float64:
         return result
 
 
-def grade_entropy(X: np.ndarray, n_jobs=os.cpu_count()) -> np.float64:
+def grade_entropy(X, normalize=True):
     '''
-    Computes grade entropy for a strict product order
+    Computes a grade entropy for a strict product order
     on the row space points.
-
-    A strict product order on embeddings of the real
-    numbers gives a partial order. On a particular data
-    set it may be desirable to quantify the extent to
-    which there is a total order vs noorder at all.
-
-    Under such an order, two points :math:`\\vec{x}, \\vec{y} \\in \\mathbb{R}^n`
-    hold the relation :math:`\\vec{x} < \\vec{y}` if-and-only-if
-    :math:`x_j < y_j\\ \\forall j \\in \\{1, \\cdots, n\\}`.
-
+    
     This function computes
 
     .. math::
-        H_g = \\frac{-\\sum_{i=1}^{m} p \\circ g(x_i) \\ln (p \\circ g(x_i))}{\\ln{m}}
+        H_g = \\frac{-\\sum_{i=1}^{k} p (g_i) \\ln (p (g_i))}{\\ln{m}}
 
     where :math:`p` is a probability distribution over the grades :math:`g`
     of the point :math:`x_i` among the indexed set of points :math:`i \\in \\{1, \\cdots, m\\}`
@@ -123,15 +156,86 @@ def grade_entropy(X: np.ndarray, n_jobs=os.cpu_count()) -> np.float64:
     --------
     >>> import numpy as np
     >>> data = np.arange(100).reshape(10,10)
-    >>> grade_entropy(data)
+    >>> pseudograde_entropy(data)
     1.0
+    '''
+    D = nx.DiGraph()
+    for i, xi in enumerate(X):
+        for j, xj in enumerate(X):
+            if np.all(xi < xj):
+                D.add_edge(j, i)
 
-    .. warning::
-        The underlying tqdm tool that achieves parallel processing will also print a progress bar. This printout can be needlessly obscure in certain environments such as IDLE. The progress bar should print normally from a BASH environment.
+    assert nx.is_directed_acyclic_graph(D), 'Graph is not a DAG.'
 
+    D = nx.transitive_reduction(D)
+    grades = {}
+    current_grade = 1
+    for node in nx.topological_sort(D):
+        if grades == {}:
+            grades[current_grade] = [node]
+        else:
+            for visited_node in grades[current_grade]:
+                if node in nx.descendants(D, visited_node):
+                    current_grade += 1
+                    grades[current_grade] = [node]
+                    break
+                else:
+                    continue
+            else:
+                grades[current_grade].append(node)
+
+    total = len(D.nodes())
+    grade_dist = [len(j) / total for i,j in grades.items()]
+    grade_dist = np.array(grade_dist)
+    result  = - np.sum(grade_dist * np.log2(grade_dist))
+    if normalize:
+        result /= np.log2(total)
+        return result
+    else:
+        return result
+
+def pseudograde_entropy(X: np.ndarray, n_jobs=1) -> np.float64:
+    '''
+    Computes a pseudograde entropy for a strict product order
+    on the row space points.
+
+    This function computes
+
+    .. math::
+        H_g = \\frac{-\\sum_{i=1}^{k} p (g_i) \\ln (p (g_i))}{\\ln{m}}
+
+    where :math:`p` is a probability distribution over the pseudogrades :math:`g`
+    of the point :math:`x_i` among the indexed set of points :math:`i \\in \\{1, \\cdots, m\\}`
+    according to a strict product order relation.
+
+    Parameters
+    ----------
+    X : array-like
+        An m x n data matrix.
+
+    Returns
+    -------
+    entropy : np.float64
+        Pseudograde entropy of product order.
+
+    References
+    ----------
+    .. [1] "Graded Poset.", https://en.wikipedia.org/wiki/Graded_poset
+    .. [2] "Entropy (Information Theory).", https://en.wikipedia.org/wiki/Entropy_(information_theory)
+    .. [3] "Partially Ordered Set.", https://en.wikipedia.org/wiki/Partially_ordered_set
+    .. [4] "Partially Ordered Set.", https://mathworld.wolfram.com/PartiallyOrderedSet.html
+    .. [5] "Product Order.", https://en.wikipedia.org/wiki/Product_order
+    .. [6] "Ranked Poset.", https://en.wikipedia.org/wiki/Ranked_poset
+    .. [7] "Total Order.", https://en.wikipedia.org/wiki/Total_order
+
+    Examples
+    --------
+    >>> import numpy as np
+    >>> data = np.arange(100).reshape(10,10)
+    >>> pseudograde_entropy(data)
+    1.0
     '''
     P = np.zeros((X.shape[0], X.shape[0]))
-    @jit(nopython=True, fastmath=True) # does fastmath actually improve speed here?
     def grade(q):
         v = np.zeros(X.shape[0])
         for j in range(X.shape[0]):
@@ -139,18 +243,78 @@ def grade_entropy(X: np.ndarray, n_jobs=os.cpu_count()) -> np.float64:
                 v[j] += 1
         return v
 
-    pool = ProcessPool(nodes=n_jobs)
-    for i, v in enumerate(pool.map(grade, range(X.shape[0]))):
-        P[:, i] = v
+    if n_jobs == 1:
+
+        for i in tqdm.tqdm(range(X.shape[0])):
+            P[:, i] = grade(i)
+
+    else:
+        try:
+            pool = ProcessPool(nodes=n_jobs)
+            for i, v in enumerate(pool.map(grade, range(X.shape[0]))):
+                P[:, i] = v
+            pool.close()
+            pool.join()
+        except ValueError:
+            pool.restart()
+            pool = ProcessPool(nodes=n_jobs)
+            for i, v in enumerate(pool.map(grade, range(X.shape[0]))):
+                P[:, i] = v
+            pool.close()
+            pool.join()
+
 
     bins, counts = np.unique(np.sum(P, axis=0), return_counts=True)
     probs = counts / np.sum(counts)
     entropy = np.sum(probs * np.log(probs)) / (np.log(1/np.sum(counts)))
     return entropy
 
-def minkowski_deviation(x: np.ndarray, order=2) -> np.float64:
+def median_correlation(X: np.ndarray, transform=lambda x: x - np.median(x, axis=0)) -> np.float64:
     '''
-    Calculates the Minkowski deviation of order p.
+    Median (multilinear) correlation.
+
+    The function estimates
+
+    .. math::
+        R_{\\mathcal{M}} \\left[ X_1, \\cdots, X_n \\right] = \\frac{\\mathcal{M} \\left[ \\prod_{j=1}^{n} \\left( X_j - \\mathcal{M}[X_j] \\right) \\right]}{\\prod_{j=1}^{n} \\sqrt[n]{\\mathcal{M}\\left[ |X_n - \\mathcal{M}[X_j]|^n \\right]}}
+
+    Parameters
+    ----------
+    X : array-like[np.float64]
+        An m x n data matrix.
+    transform : function
+        A data transform before computing coefficient.
+
+    Returns
+    -------
+    r : np.float64
+        The calculated median correlation coefficient.
+
+    References
+    ----------
+       .. [1] "Median", https://en.wikipedia.org/wiki/Median
+
+    Examples
+    ----------
+    >>> import numpy as np
+    >>> data = np.arange(100).reshape(10,10)
+    >>> median_correlation(data)
+    0.9999999999999982
+    '''
+    X = transform(X)
+    numerator = np.prod(X, axis=1)
+    numerator = np.median(numerator)
+    denominator = np.abs(X)
+    denominator = np.power(denominator, X.shape[1])
+    denominator = np.median(denominator, axis=0)
+    denominator = np.prod(denominator)
+    denominator = np.power(denominator, 1 / X.shape[1])
+    r = numerator / denominator
+    return r
+    
+def nightingale_deviation(x: np.ndarray, p=2) -> np.float64:
+    '''
+    Calculates the Nightingale deviation of order p.
     When the order = 2, it is the same as the standard
     deviation.
 
@@ -162,6 +326,7 @@ def minkowski_deviation(x: np.ndarray, order=2) -> np.float64:
     Parameters
     ----------
     x : array-like.
+        Instances of a variable.
 
     Returns
     -------
@@ -180,13 +345,11 @@ def minkowski_deviation(x: np.ndarray, order=2) -> np.float64:
     >>> import numpy as np
     >>> data = np.arange(10)
     >>> minkowski_deviation(data)
-    5.338539126015656
+    2.8722813232690143
     '''
     result = x - np.mean(x)
-    result = np.abs(result)
-    result = np.power(result, order)
-    result = np.mean(result)
-    result = np.power(result, 1 / order)
+    result = pnorm(result, p=p)
+    result *= np.power(1 / x.size, 1 / p)
     return result
 
 def reflective_correlation(X: np.ndarray) -> np.float64:
@@ -257,7 +420,7 @@ def pearson_correlation(X: np.ndarray) -> np.float64:
 
     References
     ----------
-       https://en.wikipedia.org/wiki/Pearson_correlation_coefficient
+       .. [1] "Pearson product-moment correlation coefficient.", https://en.wikipedia.org/wiki/Pearson_correlation_coefficient
 
     Examples
     ----------
@@ -490,12 +653,12 @@ def misiak_correlation(x: np.ndarray, y: np.ndarray, X: np.ndarray) -> np.float6
     result = numerator / denominator
     return result
 
-def multisemideviation(X: np.ndarray, p=1) -> np.float64:
+def nightingale_covariance(X: np.ndarray, p=1) -> np.float64:
     '''
-    This function calculates the multisemideviation
+    This function calculates the Nightingale covariance
     which is the multisemimetric between a collection
     of random variables from their expectations. The
-    multisemimetric is induced by the multiseminorm,
+    multisemimetric is induced by a multiseminorm,
     which is a generalization of the notion of a
     seminorm.
 
@@ -507,12 +670,12 @@ def multisemideviation(X: np.ndarray, p=1) -> np.float64:
     Returns
     ----------
         : np.float64
-            Multisemideviation score.
+            Nightingale covariance.
 
     See Also
     --------
     numpy.std : Standard deviation.
-    minkowski_deviation : Minkowski deviation of order p.
+    nightingale_deviation : Nightingale's deviation of order p.
 
     References
     ----------
@@ -523,16 +686,55 @@ def multisemideviation(X: np.ndarray, p=1) -> np.float64:
     --------
     >>> import numpy as np
     >>> data = np.arange(100).reshape(10,10)
-    >>> multisemideviation(data)
+    >>> nightingale_covariance(data)
     7381024072265624.0
     '''
     result = X - np.mean(X, axis=0)
-    result = np.prod(result, axis=1)
-    result = np.power(result, p)
-    result = np.mean(result)
-    result = np.power(result, 1/p)
+    result = pnorm(result, p=p)
+    result *= np.power(1 / X.shape[0], 1 / p)
     return result
 
+def nightingale_correlation(X: np.ndarray, p=1, alphas=None):
+    '''
+    Calculates the Nightingale correlation which is a
+    normalized Nightingale covariance onto the interval
+    of [0,1].
+
+    Paramaters
+    ----------
+        X: array-like
+            m x n data matrix
+
+    Returns
+    -------
+        : np.float64
+            Nightingale correlation
+    
+    See Also
+    --------
+    nightingale_deviation : Nightingale's deviation of order p.
+    nightingale_covariance : Nightingale's covariance of order p.
+
+    References
+    --------
+        .. _Seminorm(Mathworld): https://mathworld.wolfram.com/Seminorm.html
+        .. Seminorm(Wikipedia): https://en.wikipedia.org/wiki/Seminorm
+    
+
+    '''
+    if not alphas:
+        alphas = np.ones(X.shape[1]) / X.shape[1]
+    else:
+        assert np.sum(alphas) == 1.0, 'Reciprocals of alphas must sum to one.'
+        assert np.all(alphas > 1), 'Alphas must be greater than one.'
+    
+    denominator = nightingale_covariance(X, p=p)
+    numerator = X - np.mean(X, axis=0)
+    numerator = [nightingale_deviation(numerator[:,j])\
+                 for j in range(X.shape[1])]
+    numerator = np.prod(numerator)
+    return numerator / denominator
+
 def signum_correlation(X: np.ndarray) -> np.float64:
     '''
     Signum correlation coefficient.
@@ -540,7 +742,7 @@ def signum_correlation(X: np.ndarray) -> np.float64:
     This function estimates
 
     .. math::
-        R_{\\{sign}} \\left[ X_1, \\cdots, X_n \\right] = \\frac{\\mathbb{E} \\left[ \\prod_{j=1}^{n} \\text{sign} \\left( X_j - \\mathbb{E}[X_j] \\right) \\right]}{\\prod_{j=1}^{n} \\sqrt[n]{\\mathbb{E}\\left[ |\\text{sign} \\left( X_j - \\mathbb{E}[X_j] \\right)|^n \\right]}}
+        R_{\\text{sign}} \\left[ X_1, \\cdots, X_n \\right] = \\frac{\\mathbb{E} \\left[ \\prod_{j=1}^{n} \\text{sign} \\left( X_j - \\mathbb{E}[X_j] \\right) \\right]}{\\prod_{j=1}^{n} \\sqrt[n]{\\mathbb{E}\\left[ |\\text{sign} \\left( X_j - \\mathbb{E}[X_j] \\right)|^n \\right]}}
 
     Parameters
     ----------
@@ -717,3 +919,49 @@ def weak_inner_correlation():
     '''
     raise NotImplementedError
 
+
+def partial_agnesian(X: np.ndarray, t=None, k=0):
+    '''
+    Computes the partial Agnesian of order k on a
+    data matrix. If a vector of parameters is not
+    provided, then a parameter step size of unity
+    is assumed.
+
+    Parameters
+    ----------
+    X : array-like (2D)
+        m x n data matrix
+    t : array-like (1D)
+        Vector parameters corresponding to the rows of X.
+    k : Non-negative int
+        Non-negative order of the partial Agnesian operator.
+    
+    Returns
+    -------
+    : array-like[float]
+        Sequence of partial Agnesian scores.
+        
+    Examples
+    --------
+    >>> import numpy as np
+    >>> data = np.arange(5*3).reshape(5,3)
+    >>> partial_agnesian(X,k=1)
+    array([27, 27, 27, 27])
+    >>> partial_agnesian(X, k=2)
+    array([0, 0, 0])
+    >>> t = np.linspace(0, 10, 5)
+    >>> partial_agnesian(X, k=1, t=t)
+    array([1.728, 1.728, 1.728, 1.728])
+    '''
+    
+    result = np.diff(X, n=k, axis=0)
+
+    if t is not None:
+        assert t.ndim == 1, 'Parameter t must be one-dimensional.'
+        dt = np.diff(t, n=k)
+        result = result / dt[:, None]
+    
+    result = np.prod(result, axis=1)
+    return result
+
+    
